"""TCRUtils Discord. Contains various features related to bot development on discord and other stuff maybe."""


from .src import tcrd_types as types
from .src.tcrd_alias import get_guild_count
from .src.tcrd_commands import get_slash_command_ids
from .src.tcrd_constants import DISCORD_EPOCH
from .src.tcrd_dpy import escape_markdown, escape_mentions, remove_markdown
from .src.tcrd_embeds import embed, modal
from .src.tcrd_limits import DiscordLimits
from .src.tcrd_markdown import codeblock, codeblocks, discord_error, uncodeblock
from .src.tcrd_permissions import PERMISSIONS_DICT, Permission, permissions
from .src.tcrd_permissions import Permission as DiscordPermission
from .src.tcrd_permissions import permissions as discord_permissions
from .src.tcrd_shorts import confirm
from .src.tcrd_snowflake import is_snowflake
from .src.tcrd_string import IFYs, get_token
from .version import __version__



from collections.abc import Callable, Coroutine, Iterable
from functools import partial
from typing import Any

import hikari
import miru
import miru.view

from . import tcrd_types as dtypes
from .tcrd_embeds import embed, modal


class UpdatableView(miru.View):
  async def update(self) -> None:
    await self.message.edit(components=self)


class DisableableView(UpdatableView):
  async def on_timeout(self) -> None:
    await self.disable(edit=True, stop=False)
    return await super().on_timeout()

  async def disable(self, *, edit: bool = True, stop: bool = True) -> None:  # cripple_youth()
    """Set `disabled` of all children to True. Optionally edit the message with new view and/or stop the view listener for interactions."""
    for ch in self.children:
      ch.disabled = True

    if edit:
      await self.update()

    if stop:
      self.stop()


class CallbackedButton(miru.Button):
  __init_kwargs: dict
  __callback: Callable[['CallbackedButton', miru.ViewContext], Coroutine[Any, Any, None]]

  def __init_subclass__(cls, **kwargs) -> None:
    cls.__init_kwargs = kwargs

  def __init__(self, callback: Callable[['CallbackedButton', miru.ViewContext], Coroutine[Any, Any, None]], /, *, disable_on_click: bool = True, **kwargs) -> None:
    self.__callback = callback
    self.__disable_on_click = disable_on_click
    super().__init__(**{**self.__init_kwargs, **kwargs})

  async def callback(self, ctx: miru.ViewContext) -> None:
    if self.__disable_on_click:
      await ctx.view.disable(edit=True, stop=True) # type: ignore
    await self.__callback(self, ctx)


class YesButton(CallbackedButton, label='Yes', style=hikari.ButtonStyle.SUCCESS): ...


class NoButton(CallbackedButton, label='No', style=hikari.ButtonStyle.DANGER): ...


class MaybeButton(CallbackedButton, label='Maybe', style=hikari.ButtonStyle.PRIMARY): ...


async def confirm(
  responder: dtypes.HikariResponder,
  miru_client: miru.Client,
  *,
  yes_callback: Callable[[miru.Button, miru.ViewContext], Coroutine[Any, Any, None]],
  no_callback: Callable[[miru.Button, miru.ViewContext], Coroutine[Any, Any, None]],
  buttons: Iterable[miru.Button | bool] = (True, False),
  disable_on_click: bool = True,
  responder_kwargs: dtypes.HikariDictMessage = {'content': 'Please choose an option'},  # noqa: B006
  view_kwargs: dict = {'timeout': 2 * 60},  # noqa: B006
) -> hikari.Message:
  """Send a rather simple confirmator for user to click yes/no."""

  buttons = [YesButton(yes_callback, disable_on_click=disable_on_click) if x is True else x for x in buttons]
  buttons = [NoButton(no_callback, disable_on_click=disable_on_click) if x is False else x for x in buttons]

  view = DisableableView(**view_kwargs)

  for btn in buttons:
    view = view.add_item(btn) # type: ignore

  m = await responder(components=view, **responder_kwargs) # type: ignore
  miru_client.start_view(view, bind_to=m)

  return m



from collections.abc import Callable

from tcrutils import BACKTICKS, NEWLINE, cut_at, extract_error, extract_traceback

from .tcrd_limits import DiscordLimits


def codeblock(
  text: str,
  *,
  langcode: str = '',
  max_length: int = DiscordLimits.Message.LENGTH_SAFE,
  cut_at_func: Callable[[str, int], str] = cut_at,
  smart_empty: bool = True,
) -> str:
  """Pack text into a Discord codeblock.

  Note: May return str with len > max_length if max_length is ridicously low such that not even `'''\\nlangcode'''` fits

  Args:
    text: The text to put in the codeblock body.
    langcode: The language code to use in discord codeblock, for example "py".
    max_length: The maximum length of the output str.
    cut_at_func: A Callable[[str, int], str] which's returned string is of or less than the passed in int in length.
    smart_empty: If true, if `text` is '' (an empty string) it will be changed to ' ' (a space). This prevents the codeblock to think langcode is the body of the codeblock if no text is supplied.
  """
  if smart_empty and text == '':
    text = ' '

  maxlen = (
    max_length - (len(BACKTICKS) + len(langcode) + len(NEWLINE) + len(BACKTICKS)) if max_length != -1 else 9999999999999999  # Eh...
  )
  return BACKTICKS + langcode + NEWLINE + cut_at_func(text, maxlen) + BACKTICKS


def uncodeblock(text: str) -> str:
  """*Try to* remove any discord codeblocks along with their langcodes if any are found."""
  if text[-3:] == BACKTICKS and text[:3] == BACKTICKS:
    code_start = 3
    code_end = -3
    if NEWLINE in text[3:]:  # Check if there is a language code specified
      code_start = text.index(NEWLINE) + 1
    return text[code_start:code_end].strip()
  return text


def codeblocks(
  text1: str,
  *texts: str,
  langcodes: tuple[str] | None = None,
  max_length: int = DiscordLimits.Message.LENGTH_SAFE,
  cut_at_func: Callable[[str, int], str] = cut_at,
  smart_empty: bool = True,
):
  """Mutliple-codeblock(). May suffer simillar issues as `codeblock`, like the total_max_length not actually being max because codeblock() may return something slightly longer.

  For more info read codeblock() docstring
  """
  texts = text1, *texts

  if langcodes is None:
    langcodes = ('',) * len(texts)

  out = ''

  for text, langcode in zip(texts, langcodes, strict=True):
    out += codeblock(text, langcode=langcode, max_length=max_length - len(out), cut_at_func=cut_at_func, smart_empty=smart_empty)

  return out


def discord_error(e: BaseException) -> str:
  return codeblocks(
    extract_error(e),
    extract_traceback(e),
    langcodes=(
      'txt',
      'py',
    ),
    max_length=DiscordLimits.Message.LENGTH_SAFEST,
    smart_empty=True,
  )



from collections.abc import Callable, Iterable

import hikari

Permission = hikari.Permissions

PERMISSIONS_DICT = {x: int(y) for x, y in hikari.Permissions.__dict__.items() if not x.startswith('_') and x == x.upper() and x not in ['MANAGE_EMOJIS_AND_STICKERS']}


class _Permissions:
  """### Contains useful utilities related to checking discord permissions.

  This is tested with the [Hikari](https://pypi.org/project/hikari/) module and most likely will not work with any module that uses a different object structure to Hikari, although `has()` function will work with any module since it takes solely the permission ints.

  use `has*` (`has`, `has_by_GMCE`, etc.) functions to return bool based on whether the user matches the selected permissions.
  use `perms.devlist = (1234,)` to set a list of devs. This step is entirely optional although you will not be able to use `allow_devs=True` option in some functions
  pass in any method you want to use for comparing permissions for example `method=perms.ANY` or `method=perms.ALL`

  It's recommended to import the permissions object as follows:
  ```py
  from tcrutils.discord import permissions as perms
  # or any other short name you wish to use, although that's the name that will be used in many explainations and examples
  ```
  """

  _devs: tuple | None = None

  @property
  def devlist(self) -> list:
    return list(self._devs) # type: ignore

  @devlist.setter
  def devlist(self, x: Iterable[int]) -> None:
    if isinstance(x, list | tuple | set) and all(isinstance(item, int) for item in x):
      self._devs = tuple(x)

  @devlist.deleter
  def devlist(self) -> None:
    self._devs = None

  @staticmethod
  def ALL(possessed: int, required: int, /) -> bool:
    """Used in `has*()` functions (`has()`, `has_by_GMCE()`, etc.) as the method=ALL argument.

    You may, although it's not recommended to use it by itself.

    Equivalent to:
    ```py
    return (possessed & required) == required # Checks if all bits of required are contained in possessed.
    ```
    """
    return (possessed & required) == required

  @staticmethod
  def ANY(possessed: int, required: int, /) -> bool:
    """Used in `has*()` functions (`has()`, `has_by_GMCE()`, etc.) as the method=ANY argument.

    You may, although it's not recommended to use it by itself.

    Equivalent to:
    ```py
    return bool(possessed & required) # Checks if any of the bits of required are set in possessed.
    ```
    """
    return bool(possessed & required)

  def has(
    self,
    possessed: int,
    required: int,
    method: Callable[[int, int], bool] = ALL,
    *,
    allow_administrator=False,
  ) -> bool:
    """Return bool whether `possessed` shares `ALL` (or `ANY` if that method is selected) bits or more with `required`.

    `allow_administrator`: Will ignore method and required permissions and return True if possessed permissions have the admininistrator bit set.
    """
    if allow_administrator and (possessed & Permission.ADMINISTRATOR):
      return True

    return method(possessed, required)

  def has_by_GMCE(
    self,
    event: hikari.GuildMessageCreateEvent,
    required: int,
    *,
    method: Callable[[int, int], bool] = ALL,
    allow_administrator=False,
    allow_owner=False,
    allow_dev=False,
  ) -> bool:
    if allow_dev:
      if self._devs is not None:
        if event.author.id in self._devs:
          return True
      else:
        raise ValueError("You can't use allow_dev=True without setting up devs first with `perms.devlist = [1337, 1234] # A list of ints (discord user ids)`")

    if allow_owner and event.author.id == event.get_guild().owner_id: # type: ignore
      return True

    roles = event.message.member.get_roles() # type: ignore

    if allow_administrator and any(role.permissions & hikari.Permissions.ADMINISTRATOR for role in roles):  # fmt: skip
      return True

    return any(method(role.permissions, required) for role in roles)

  def to_str(self, permissions: int) -> str:
    if not permissions:
      return f'`No permissions`'

    return ', '.join([f'`{name.replace("_", " ").title()}`' for name, value in PERMISSIONS_DICT.items() if value & permissions])


permissions = _Permissions()



import arc
import hikari

from .tcrd_types import CommandIDsDict


def _get_pair_for_command(cmd: arc.SlashCommand | arc.SlashSubCommand | arc.SlashSubGroup, *previous_names):
  if hasattr(cmd, '_instances'):
    id = cmd.instances[None].id if None in cmd.instances else next(iter(cmd.instances.values())).id  # type: ignore
    return (' '.join([*previous_names, cmd.name][::-1]), id)

  return _get_pair_for_command(cmd.parent, *previous_names, cmd.name) # type: ignore

def get_slash_command_ids(arc_client: arc.GatewayClient):
  commands = list(arc_client.walk_commands(hikari.CommandType.SLASH, callable_only=True))

  pairs = [_get_pair_for_command(cmd) for cmd in commands]

  return CommandIDsDict(pairs)



import pathlib as p
import re as regex
import sys

import setuptools

PACKAGE_NAME = 'tcrdiscord'

if __name__ == '__main__':
  version_file = p.Path(f'./{PACKAGE_NAME}/version.py')
  version_pattern = r"__version__\s*=\s*[\"'](.*?)[\"']"

  __version__ = regex.search(version_pattern, version_file.read_text()).group(1)

  setuptools.setup(
    name=PACKAGE_NAME,
    version=__version__,
    description='Various utilities for working with the hikari library and the Discord API.',
    long_description=p.Path('./README.md').read_text(),
    long_description_content_type='text/markdown',
    url=f'https://github.com/anamoyee/{PACKAGE_NAME}',
    author='anamoyee',
    license='GPL-3.0 license',
    project_urls={
      'Source': 'https://github.com/anamoyee/tcrdiscord',
    },
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 3.11',
      'Programming Language :: Python :: 3.12',
      'Topic :: Utilities',
    ],
    python_requires='>=3.11,<3.13',
    install_requires=p.Path('./requirements.txt').read_text().strip().split(),
    packages=setuptools.find_packages(),
    include_package_data=True,
  )



def is_snowflake(snowflake: int | str, *, allow_string=True) -> bool:
  if not isinstance(snowflake, ((int | str) if allow_string else int)):
    return False

  if not str(snowflake).isnumeric() or (int(snowflake) not in range(1 << 64)):
    return False

  return True



import collections.abc as _cabc
import typing as _t
from collections.abc import Awaitable as __Awaitable
from collections.abc import Callable as __Callable
from collections.abc import Generator as _Generator
from typing import Unpack as __Unpack

import hikari as _hk

from .tcrd_string import IFYs as _IFYs


class HikariDictMessage(_t.TypedDict):
  """Signature of some hikari respond() functions."""

  content: str
  attachment: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.files.Resourceish]]
  attachments: _t.NotRequired[_hk.undefined.UndefinedOr[_cabc.Sequence[_hk.files.Resourceish]]]
  component: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.api.CommandBuilder]]
  components: _t.NotRequired[_hk.undefined.UndefinedOr[_cabc.Sequence[_hk.api.ComponentBuilder]]]
  embed: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.Embed]]
  embeds: _t.NotRequired[_hk.undefined.UndefinedOr[_cabc.Sequence[_hk.Embed]]]
  sticker: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.snowflakes.SnowflakeishOr[_hk.PartialSticker]]]
  stickers: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.snowflakes.SnowflakeishSequence[_hk.PartialSticker]]]
  tts: _t.NotRequired[_hk.undefined.UndefinedOr[bool]]
  reply: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.snowflakes.SnowflakeishOr[_hk.PartialMessage]]]
  reply_must_exist: _t.NotRequired[_hk.undefined.UndefinedOr[bool]]
  mentions_everyone: _t.NotRequired[_hk.undefined.UndefinedOr[bool]]
  mentions_reply: _t.NotRequired[_hk.undefined.UndefinedOr[bool]]
  user_mentions: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.snowflakes.SnowflakeishSequence[_hk.users.PartialUser] | bool]]
  role_mentions: _t.NotRequired[_hk.undefined.UndefinedOr[_hk.snowflakes.SnowflakeishSequence[_hk.guilds.PartialRole] | bool]]
  flags: _t.NotRequired[_hk.undefined.UndefinedType | (int | _hk.MessageFlag)]


HikariResponder = __Callable[[__Unpack[HikariDictMessage]], __Awaitable[_hk.Message]]


hikari_dict_message_defaults = {
  'attachment': _hk.UNDEFINED,
  'attachments': _hk.UNDEFINED,
  'component': _hk.UNDEFINED,
  'components': _hk.UNDEFINED,
  'embed': _hk.UNDEFINED,
  'embeds': _hk.UNDEFINED,
  'sticker': _hk.UNDEFINED,
  'stickers': _hk.UNDEFINED,
  'tts': _hk.UNDEFINED,
  'reply': _hk.UNDEFINED,
  'reply_must_exist': _hk.UNDEFINED,
  'mentions_everyone': _hk.UNDEFINED,
  'mentions_reply': _hk.UNDEFINED,
  'user_mentions': _hk.UNDEFINED,
  'role_mentions': _hk.UNDEFINED,
  'flags': _hk.UNDEFINED,
}

class CommandIDsDict(dict):
  """Mapping of command_name -> command_id for slash commands in Discord."""

  def mentions(self) -> _Generator[tuple[int, str], None, None]:
    """Iterate over `tuple[command_id, command_mention]` of commands in this dict where command mention is `f'</{command_name}:{command_id}>'`."""

    for command_name, command_id in self.items():
      yield command_id, _IFYs.commandify(command_name, command_id)

  def mentions_named(self) -> _Generator[tuple[str, str], None, None]:
    """Iterate over `tuple[command_name, command_mention]` of commands in this dict where command mention is `f'</{command_name}:{command_id}>'`."""

    for command_name, command_id in self.items():
      yield command_name, _IFYs.commandify(command_name, command_id)
